------------------------------------------------------------------------------
--- A library to support the type-oriented construction of Web User Interfaces
--- (WUIs).
---
--- The ideas behind the application and implementation of WUIs are
--- described in the companion paper "Type-Oriented Construction of
--- Web User Interfaces".
---
--- @author Michael Hanus
--- @version March 2006
------------------------------------------------------------------------------

module WUI(--WuiState,cgiRef2state,state2cgiRef,value2state,state2value,
           --states2state,state2states,altstate2state,state2altstate,
           Rendering,WuiSpec,
           withRendering,withError,withCondition,adaptWSpec,transformWSpec,
           wHidden,wConstant,wInt,
           wString,wStringSize,wRequiredString,wRequiredStringSize,wTextArea,
           wMaybeString,
           wSelect,wSelectBool,wRadioSelect,wRadioBool,
           wPair,wTriple,w4Tuple,w5Tuple,w9Tuple,w10Tuple,w11Tuple,
           wList,wListWithHeadings,wHList,wMatrix,wEither,
           mainWUI,wui2html,wuiInForm,wuiWithErrorForm)
 where

import HTML
import Read(readNat)
import List(elemIndex)
import Maybe
import Char(isDigit,isSpace)
import ReadShowTerm

infixl 0 `withRendering`
infixl 0 `withError`
infixl 0 `withCondition`

------------------------------------------------------------------------------
--- An internal WUI state is used to maintain the cgi references of the input
--- fields as a structure that corresponds to the structure of the edit data.
data WuiState =
     Leaf CgiRef            -- reference to elementary input field
   | Hidden String          -- string representation of a hidden value
   | CompNode [WuiState]    -- composition of trees (substructures)
   | AltNode (Int,WuiState) -- alternative of trees (union of substructures)

cgiRef2state :: CgiRef -> WuiState
cgiRef2state cr = Leaf cr

state2cgiRef :: WuiState -> CgiRef
state2cgiRef (Leaf cr) = cr

value2state :: _ -> WuiState
value2state v = Hidden (showQTerm v)

state2value :: WuiState -> _
state2value (Hidden s) = readQTerm s

states2state :: [WuiState] -> WuiState
states2state sts = CompNode sts

state2states :: WuiState -> [WuiState]
state2states (CompNode sts) = sts

altstate2state :: (Int,WuiState) -> WuiState
altstate2state alt = AltNode alt

state2altstate :: WuiState -> (Int,WuiState)
state2altstate (AltNode alt) = alt

------------------------------------------------------------------------------
--- A rendering is a function that combines the visualization of components
--- of a data structure into some HTML expression.
type Rendering = [HtmlExp] -> HtmlExp

--- WuiParams specify the parameters of an individual Wui component type:
--- * the standard rendering
--- * an error message shown in case of illegal inputs
--- * a condition to specify legal input values
type WuiParams a = (Rendering, String, a->Bool)

renderOf (render,_,_) = render

------------------------------------------------------------------------------
--- The type of WUI specifications.
--- The first component are parameters specifying the behavior of this WUI type
--- (rendering, error message, and constraints on inputs).
--- The second component is a "show" function returning an HTML expression for
--- the edit fields and a WUI state containing the CgiRefs to extract
--- the values from the edit fields.
--- The third component is "read" function to extract the values from
--- the edit fields for a given cgi environment (returned as (Just v)).
--- If the value is not legal, Nothing is returned. The second component
--- of the result contains an HTML edit expression
--- together with a WUI state to edit the value again.
data WuiSpec a =
  WuiSpec (WuiParams a)
          (WuiParams a -> a -> (HtmlExp,WuiState))
          (WuiParams a -> CgiEnv -> WuiState -> (Maybe a,(HtmlExp,WuiState)))

--- Puts a new rendering function into a WUI specification.
withRendering :: WuiSpec a -> Rendering -> WuiSpec a
withRendering (WuiSpec (_,errmsg,legal) showhtml readvalue) render =
  WuiSpec (render,errmsg,legal) showhtml readvalue


--- Puts a new error message into a WUI specification.
withError :: WuiSpec a -> String -> WuiSpec a
withError (WuiSpec (render,_,legal) showhtml readvalue) errmsg =
  WuiSpec (render,errmsg,legal) showhtml readvalue

--- Puts a new condition into a WUI specification.
withCondition :: WuiSpec a -> (a -> Bool) -> WuiSpec a
withCondition (WuiSpec (render,errmsg,_) showhtml readvalue) legal =
              (WuiSpec (render,errmsg,legal) showhtml readvalue)

--- Transforms a WUI specification from one type to another.
transformWSpec :: (a->b,b->a) -> WuiSpec a -> WuiSpec b
transformWSpec (a2b,b2a) (WuiSpec wparamsa showhtmla readvaluea) =
  WuiSpec (transParam b2a wparamsa)
          (\wparamsb b -> showhtmla (transParam a2b wparamsb) (b2a b))
          (\wparamsb env wst ->
            let (mba,errv) = readvaluea (transParam a2b wparamsb) env wst
             in (maybe Nothing (Just . a2b) mba, errv))
 where
  transParam :: (b->a) -> WuiParams a -> WuiParams b
  transParam toa (render,errmsg,legal) = (render,errmsg,legal . toa)

--- Adapt a WUI specification to a new type. For this purpose,
--- the first argument must be a transformation mapping values
--- from the old type to the new type. This function must be bijective
--- and operationally invertible (i.e., the inverse must be computable
--- by narrowing). Otherwise, use <code>transformWSpec</code>!
adaptWSpec :: (a->b) -> WuiSpec a -> WuiSpec b
adaptWSpec a2b = transformWSpec (a2b,invert a2b)

-- Compute the inverse of a function by exploiting function patterns:
invert :: (a->b) -> b -> a
invert f = f_invert
 where
  local_f x = f x
  f_invert (local_f x) = x  -- here we use a function pattern
  --f_invert y | (local_f x) =:<= y = x  where x free


------------------------------------------------------------------------------
-- A collection of basic WUIs and WUI combinators:

--- A hidden widget for a value that is not shown in the WUI.
--- Usually, this is used in components of larger
--- structures, e.g., internal identifiers, data base keys.
wHidden :: WuiSpec a
wHidden =
  WuiSpec (head,"?",const True) -- dummy values, not used
          (\_ v -> (hempty, value2state v))
          (\_ _ s -> (Just (state2value s), (hempty,s)))

--- A widget for values that are shown but cannot be modified.
--- The first argument is a mapping of the value into a HTML expression
--- to show this value.
wConstant :: (a->HtmlExp) -> WuiSpec a
wConstant showhtml =
  WuiSpec (head,"?",const True)
          (\wparams v -> ((renderOf wparams) [showhtml v], value2state v))
          (\(render,_,_) _ s -> let v = state2value s in
                                (Just v, (render [showhtml v],s)))

--- A widget for editing integer values.
wInt :: WuiSpec Int
wInt =
  WuiSpec (head,"Illegal integer:",const True)
          (\wparams v -> intWidget (renderOf wparams) (show v))
          (\(render,errmsg,legal) env s ->
            let input = env (state2cgiRef s)
                renderr = renderError render errmsg
             in maybe (Nothing, intWidget renderr input)
                      (\v -> if legal v
                             then (Just v,  intWidget render input)
                             else (Nothing, intWidget renderr input))
                      (readMaybeInt (stripSpaces input)))
 where
  intWidget render s = let ref free in
    (render [textfield ref s `addAttr` ("size","6")], cgiRef2state ref)

-- Remove leading and ending spaces in a string.
stripSpaces :: String -> String
stripSpaces = reverse . dropWhile isSpace . reverse . dropWhile isSpace

-- Read a (possibly negative) integer in a string.
-- Return Nothing is this is not an integer string.
readMaybeInt :: String -> Maybe Int
readMaybeInt "" = Nothing
readMaybeInt (v:s) | v=='-'  = maybe Nothing (\i->Just (-i)) (acc 0 s)
                   | isDigit v  = acc 0 (v:s)
                   | otherwise  = Nothing
 where
  acc n "" = Just n
  acc n (c:cs) | isDigit c = acc (10*n + ord c - ord '0') cs
               | otherwise = Nothing


checkLegalInput (render,errmsg,legal) value2widget value =
  if legal value
  then (Just value, value2widget render value)
  else (Nothing,    value2widget (renderError render errmsg) value)

--- A widget for editing string values.
wString :: WuiSpec String
wString = wStringAttrs []

--- A widget for editing string values with a size attribute.
wStringSize :: Int -> WuiSpec String
wStringSize size = wStringAttrs [("size",show size)]

--- A widget for editing string values with some attributes for the
--- text field.
wStringAttrs :: [(String,String)] -> WuiSpec String
wStringAttrs attrs =
  WuiSpec (head, "?", const True)
          (\wparams v -> stringWidget (renderOf wparams) v)
          (\wparams env s ->
                checkLegalInput wparams stringWidget (env (state2cgiRef s)))
 where
  stringWidget render v =
    let ref free in
    (render [foldr (flip addAttr) (textfield ref v) attrs], cgiRef2state ref)

--- A widget for editing string values that are required to be non-empty.
wRequiredString :: WuiSpec String
wRequiredString =
  wString `withError`     "Missing input:"
          `withCondition` (not . null)

--- A widget with a size attribute for editing string values
--- that are required to be non-empty.
wRequiredStringSize :: Int -> WuiSpec String
wRequiredStringSize size =
  wStringSize size `withError`     "Missing input:"
                   `withCondition` (not . null)

--- A widget for editing string values in a text area.
--- The argument specifies the height and width of the text area.
wTextArea :: (Int,Int) -> WuiSpec String
wTextArea dims =
  WuiSpec (head, "?", const True)
          (\wparams v -> textareaWidget (renderOf wparams) v)
          (\wparams env s ->
                checkLegalInput wparams textareaWidget (env (state2cgiRef s)))
 where
  textareaWidget render v = let ref free in
                            (render [textarea ref dims v], cgiRef2state ref)


--- A widget to select nothing or a string in some string widget.
--- The first and second arguments are the list of HTML expressions
--- that are shown after the Nothing and Just radio button, respectively.
--- The third argument is the string widget.
wMaybeString :: [HtmlExp] -> [HtmlExp] -> WuiSpec String ->
                WuiSpec (Maybe String)
wMaybeString nothingexp justexp (WuiSpec renders shows reads) =
 let (WuiSpec renderb showb readb) = wRadioBool nothingexp justexp in
 WuiSpec
  (renderTuple, tupleError, const True)
  (\wparams mbs ->
    let (heb,rtb) = showb renderb (mbs/=Nothing)
        (hes,rts) = shows renders (maybe "" id mbs)
     in ((renderOf wparams) [heb,hes], states2state [rtb,rts]))
  (\ (render,errmsg,legal) env s ->
    let [rb,rs] = state2states s
        (rbv,(heb,rtb)) = readb renderb env rb
        (rsv,(hes,rts)) = reads renders env rs
        errhexps = [heb,hes]
        errstate = states2state [rtb,rts]
     in if rbv==Nothing || rsv==Nothing
        then (Nothing, (render errhexps, errstate))
        else let value = if fromJust rbv
                         then Just (fromJust rsv)
                         else Nothing in
             if legal value
             then (Just value, (render errhexps, errstate))
             else (Nothing,    (renderError render errmsg errhexps, errstate)))


--- A widget to select a value from a given list of values.
--- The current value should be contained in the value list and is preselected.
--- The first argument is a mapping from values into strings to be shown
--- in the selection widget.
wSelect :: (a->String) -> [a] -> WuiSpec a
wSelect showelem selset =
  WuiSpec (head,"?",const True)
          (\wparams v -> selWidget (renderOf wparams) v)
          (\wparams env s ->
             checkLegalInput wparams selWidget
                             (selset !! readNat (env (state2cgiRef s))))
 where
  selWidget render v =
    let ref free
        idx = elemIndex v selset
        namevalues = zip (map showelem selset) (map show [0..])
     in (render [maybe (selection ref namevalues)
                       (\i -> selectionInitial ref namevalues i)
                       idx],
         cgiRef2state ref)

--- A widget to select a Boolean value.
wSelectBool :: WuiSpec Bool
wSelectBool = wSelect show [False,True]

--- A widget to select a value from a given list of values via a radio button.
--- The current value should be contained in the value list and is preselected.
--- The first argument is a mapping from values into HTML expressions
--- that are shown for each item after the radio button.
wRadioSelect :: (a->[HtmlExp]) -> [a] -> WuiSpec a
wRadioSelect showelem selset =
  WuiSpec (renderTuple, tupleError, const True)
          (\wparams v -> radioWidget (renderOf wparams) v)
          (\wparams env s ->
             checkLegalInput wparams radioWidget
                             (selset !! readNat (env (state2cgiRef s))))
 where
  radioWidget render v =
    let ref free
        idx = maybe 0 id (elemIndex v selset)
        numhitems = zip [0..] (map showelem selset)
        showItem (i,s) = table [[[(if i==idx then radio_main else radio_other)
                                        ref (show i)],s]]
     in (render (map showItem numhitems),
         cgiRef2state ref)

--- A widget to select a Boolean value via a radio button.
--- The arguments are the list of HTML expressions that are shown after
--- the true and false radio button, respectively.
wRadioBool :: [HtmlExp] -> [HtmlExp] -> WuiSpec Bool
wRadioBool falseexp trueexp =
  wRadioSelect (\b->if b then trueexp else falseexp) [False,True]

--- Widget combinator for pairs.
wPair :: WuiSpec a -> WuiSpec b -> WuiSpec (a,b)
wPair (WuiSpec rendera showa reada) (WuiSpec renderb showb readb) =
 WuiSpec
  (renderTuple, tupleError, const True)
  (\wparams (va,vb) ->
    let (hea,rta) = showa rendera va
        (heb,rtb) = showb renderb vb
     in ((renderOf wparams) [hea,heb], states2state [rta,rtb]))
  (\ (render,errmsg,legal) env s ->
    let [ra,rb] = state2states s
        (rav,(hea,rta)) = reada rendera env ra
        (rbv,(heb,rtb)) = readb renderb env rb
        errhexps = [hea,heb]
        errstate = states2state [rta,rtb]
     in if rav==Nothing || rbv==Nothing
        then (Nothing, (render errhexps, errstate))
        else let value = (fromJust rav, fromJust rbv) in
             if legal value
             then (Just value, (render errhexps, errstate))
             else (Nothing,    (renderError render errmsg errhexps, errstate)))


--- Widget combinator for triples.
wTriple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec (a,b,c)
wTriple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
          (WuiSpec renderc showc readc) =
 WuiSpec
  (renderTuple, tupleError, const True)
  (\wparams (va,vb,vc) ->
    let (hea,rta) = showa rendera va
        (heb,rtb) = showb renderb vb
        (hec,rtc) = showc renderc vc
     in ((renderOf wparams) [hea,heb,hec], states2state [rta,rtb,rtc]))
  (\ (render,errmsg,legal) env s ->
    let [ra,rb,rc] = state2states s
        (rav,(hea,rta)) = reada rendera env ra
        (rbv,(heb,rtb)) = readb renderb env rb
        (rcv,(hec,rtc)) = readc renderc env rc
        errhexps = [hea,heb,hec]
        errstate = states2state [rta,rtb,rtc]
     in if rav==Nothing || rbv==Nothing || rcv==Nothing
        then (Nothing, (render errhexps, errstate))
        else let value = (fromJust rav, fromJust rbv, fromJust rcv) in
             if legal value
             then (Just value, (render errhexps, errstate))
             else (Nothing,    (renderError render errmsg errhexps, errstate)))


--- Widget combinator for tuples of arity 4.
w4Tuple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec d -> WuiSpec (a,b,c,d)
w4Tuple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
        (WuiSpec renderc showc readc) (WuiSpec renderd showd readd) =
 WuiSpec
  (renderTuple,tupleError, const True)
  (\wparams (va,vb,vc,vd) ->
       let (hea,rta) = showa rendera va
           (heb,rtb) = showb renderb vb
           (hec,rtc) = showc renderc vc
           (hed,rtd) = showd renderd vd
        in ((renderOf wparams) [hea,heb,hec,hed],
            states2state [rta,rtb,rtc,rtd]))
  (\ (render,errmsg,legal) env s ->
      let [ra,rb,rc,rd] = state2states s
          (rav,(hea,rta)) = reada rendera env ra
          (rbv,(heb,rtb)) = readb renderb env rb
          (rcv,(hec,rtc)) = readc renderc env rc
          (rdv,(hed,rtd)) = readd renderd env rd
          errhexps = [hea,heb,hec,hed]
          errstate = states2state [rta,rtb,rtc,rtd]
       in if rav==Nothing || rbv==Nothing || rcv==Nothing || rdv==Nothing
          then (Nothing, (render errhexps, errstate))
          else let value = (fromJust rav, fromJust rbv, fromJust rcv,
                            fromJust rdv) in
               if legal value
               then (Just value,(render errhexps, errstate))
               else (Nothing,   (renderError render errmsg errhexps,errstate)))


--- Widget combinator for tuples of arity 5.
w5Tuple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec d -> WuiSpec e
           -> WuiSpec (a,b,c,d,e)
w5Tuple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
        (WuiSpec renderc showc readc) (WuiSpec renderd showd readd)
        (WuiSpec rendere showe reade) =
 WuiSpec
  (renderTuple,tupleError, const True)
  (\wparams (va,vb,vc,vd,ve) ->
       let (hea,rta) = showa rendera va
           (heb,rtb) = showb renderb vb
           (hec,rtc) = showc renderc vc
           (hed,rtd) = showd renderd vd
           (hee,rte) = showe rendere ve
        in ((renderOf wparams) [hea,heb,hec,hed,hee],
            states2state [rta,rtb,rtc,rtd,rte]))
  (\ (render,errmsg,legal) env s ->
      let [ra,rb,rc,rd,re] = state2states s
          (rav,(hea,rta)) = reada rendera env ra
          (rbv,(heb,rtb)) = readb renderb env rb
          (rcv,(hec,rtc)) = readc renderc env rc
          (rdv,(hed,rtd)) = readd renderd env rd
          (rev,(hee,rte)) = reade rendere env re
          errhexps = [hea,heb,hec,hed,hee]
          errstate = states2state [rta,rtb,rtc,rtd,rte]
       in if rav==Nothing || rbv==Nothing || rcv==Nothing || rdv==Nothing ||
             rev==Nothing
          then (Nothing, (render errhexps, errstate))
          else let value = (fromJust rav, fromJust rbv, fromJust rcv,
                            fromJust rdv, fromJust rev) in
               if legal value
               then (Just value,(render errhexps, errstate))
               else (Nothing,   (renderError render errmsg errhexps,errstate)))


--- Widget combinator for tuples of arity 9.
w9Tuple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec d -> WuiSpec e ->
           WuiSpec f -> WuiSpec g -> WuiSpec h -> WuiSpec i ->
           WuiSpec (a,b,c,d,e,f,g,h,i)
w9Tuple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
        (WuiSpec renderc showc readc) (WuiSpec renderd showd readd)
        (WuiSpec rendere showe reade) (WuiSpec renderf showf readf)
        (WuiSpec renderg showg readg) (WuiSpec renderh showh readh)
        (WuiSpec renderi showi readi) =
 WuiSpec
  (renderTuple,tupleError, const True)
  (\wparams (va,vb,vc,vd,ve,vf,vg,vh,vi) ->
       let (hea,rta) = showa rendera va
           (heb,rtb) = showb renderb vb
           (hec,rtc) = showc renderc vc
           (hed,rtd) = showd renderd vd
           (hee,rte) = showe rendere ve
           (hef,rtf) = showf renderf vf
           (heg,rtg) = showg renderg vg
           (heh,rth) = showh renderh vh
           (hei,rti) = showi renderi vi
        in ((renderOf wparams) [hea,heb,hec,hed,hee,hef,heg,heh,hei],
            states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti]))
  (\ (render,errmsg,legal) env s ->
      let [ra,rb,rc,rd,re,rf,rg,rh,ri] = state2states s
          (rav,(hea,rta)) = reada rendera env ra
          (rbv,(heb,rtb)) = readb renderb env rb
          (rcv,(hec,rtc)) = readc renderc env rc
          (rdv,(hed,rtd)) = readd renderd env rd
          (rev,(hee,rte)) = reade rendere env re
          (rfv,(hef,rtf)) = readf renderf env rf
          (rgv,(heg,rtg)) = readg renderg env rg
          (rhv,(heh,rth)) = readh renderh env rh
          (riv,(hei,rti)) = readi renderi env ri
          errhexps = [hea,heb,hec,hed,hee,hef,heg,heh,hei]
          errstate = states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti]
       in if rav==Nothing || rbv==Nothing || rcv==Nothing || rdv==Nothing ||
             rev==Nothing || rfv==Nothing || rgv==Nothing || rhv==Nothing ||
             riv==Nothing
          then (Nothing, (render errhexps, errstate))
          else let value = (fromJust rav, fromJust rbv, fromJust rcv,
                            fromJust rdv, fromJust rev, fromJust rfv,
                            fromJust rgv, fromJust rhv, fromJust riv) in
               if legal value
               then (Just value,(render errhexps,errstate))
               else (Nothing,   (renderError render errmsg errhexps,errstate)))


--- Widget combinator for tuples of arity 10.
w10Tuple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec d -> WuiSpec e ->
            WuiSpec f -> WuiSpec g -> WuiSpec h -> WuiSpec i -> WuiSpec j ->
            WuiSpec (a,b,c,d,e,f,g,h,i,j)
w10Tuple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
         (WuiSpec renderc showc readc) (WuiSpec renderd showd readd)
         (WuiSpec rendere showe reade) (WuiSpec renderf showf readf)
         (WuiSpec renderg showg readg) (WuiSpec renderh showh readh)
         (WuiSpec renderi showi readi) (WuiSpec renderj showj readj) =
 WuiSpec
  (renderTuple,tupleError, const True)
  (\wparams (va,vb,vc,vd,ve,vf,vg,vh,vi,vj) ->
       let (hea,rta) = showa rendera va
           (heb,rtb) = showb renderb vb
           (hec,rtc) = showc renderc vc
           (hed,rtd) = showd renderd vd
           (hee,rte) = showe rendere ve
           (hef,rtf) = showf renderf vf
           (heg,rtg) = showg renderg vg
           (heh,rth) = showh renderh vh
           (hei,rti) = showi renderi vi
           (hej,rtj) = showj renderj vj
        in ((renderOf wparams) [hea,heb,hec,hed,hee,hef,heg,heh,hei,hej],
            states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti,rtj]))
  (\ (render,errmsg,legal) env s ->
      let [ra,rb,rc,rd,re,rf,rg,rh,ri,rj] = state2states s
          (rav,(hea,rta)) = reada rendera env ra
          (rbv,(heb,rtb)) = readb renderb env rb
          (rcv,(hec,rtc)) = readc renderc env rc
          (rdv,(hed,rtd)) = readd renderd env rd
          (rev,(hee,rte)) = reade rendere env re
          (rfv,(hef,rtf)) = readf renderf env rf
          (rgv,(heg,rtg)) = readg renderg env rg
          (rhv,(heh,rth)) = readh renderh env rh
          (riv,(hei,rti)) = readi renderi env ri
          (rjv,(hej,rtj)) = readj renderj env rj
          errhexps = [hea,heb,hec,hed,hee,hef,heg,heh,hei,hej]
          errstate = states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti,rtj]
       in if rav==Nothing || rbv==Nothing || rcv==Nothing || rdv==Nothing ||
             rev==Nothing || rfv==Nothing || rgv==Nothing || rhv==Nothing ||
             riv==Nothing || rjv==Nothing
          then (Nothing, (render errhexps, errstate))
          else let value = (fromJust rav, fromJust rbv, fromJust rcv,
                            fromJust rdv, fromJust rev, fromJust rfv,
                            fromJust rgv, fromJust rhv, fromJust riv,
                            fromJust rjv) in
               if legal value
               then (Just value,(render errhexps,errstate))
               else (Nothing,   (renderError render errmsg errhexps,errstate)))


--- Widget combinator for tuples of arity 11.
w11Tuple :: WuiSpec a -> WuiSpec b -> WuiSpec c -> WuiSpec d -> WuiSpec e ->
            WuiSpec f -> WuiSpec g -> WuiSpec h -> WuiSpec i -> WuiSpec j ->
            WuiSpec k -> WuiSpec (a,b,c,d,e,f,g,h,i,j,k)
w11Tuple (WuiSpec rendera showa reada) (WuiSpec renderb showb readb)
         (WuiSpec renderc showc readc) (WuiSpec renderd showd readd)
         (WuiSpec rendere showe reade) (WuiSpec renderf showf readf)
         (WuiSpec renderg showg readg) (WuiSpec renderh showh readh)
         (WuiSpec renderi showi readi) (WuiSpec renderj showj readj)
         (WuiSpec renderk showk readk) =
 WuiSpec
  (renderTuple,tupleError, const True)
  (\wparams (va,vb,vc,vd,ve,vf,vg,vh,vi,vj,vk) ->
       let (hea,rta) = showa rendera va
           (heb,rtb) = showb renderb vb
           (hec,rtc) = showc renderc vc
           (hed,rtd) = showd renderd vd
           (hee,rte) = showe rendere ve
           (hef,rtf) = showf renderf vf
           (heg,rtg) = showg renderg vg
           (heh,rth) = showh renderh vh
           (hei,rti) = showi renderi vi
           (hej,rtj) = showj renderj vj
           (hek,rtk) = showk renderk vk
        in ((renderOf wparams) [hea,heb,hec,hed,hee,hef,heg,heh,hei,hej,hek],
            states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti,rtj,rtk]))
  (\ (render,errmsg,legal) env s ->
      let [ra,rb,rc,rd,re,rf,rg,rh,ri,rj,rk] = state2states s
          (rav,(hea,rta)) = reada rendera env ra
          (rbv,(heb,rtb)) = readb renderb env rb
          (rcv,(hec,rtc)) = readc renderc env rc
          (rdv,(hed,rtd)) = readd renderd env rd
          (rev,(hee,rte)) = reade rendere env re
          (rfv,(hef,rtf)) = readf renderf env rf
          (rgv,(heg,rtg)) = readg renderg env rg
          (rhv,(heh,rth)) = readh renderh env rh
          (riv,(hei,rti)) = readi renderi env ri
          (rjv,(hej,rtj)) = readj renderj env rj
          (rkv,(hek,rtk)) = readk renderk env rk
          errhexps = [hea,heb,hec,hed,hee,hef,heg,heh,hei,hej,hek]
          errstate = states2state [rta,rtb,rtc,rtd,rte,rtf,rtg,rth,rti,rtj,rtk]
       in if rav==Nothing || rbv==Nothing || rcv==Nothing || rdv==Nothing ||
             rev==Nothing || rfv==Nothing || rgv==Nothing || rhv==Nothing ||
             riv==Nothing || rjv==Nothing || rkv==Nothing
          then (Nothing, (render errhexps, errstate))
          else let value = (fromJust rav, fromJust rbv, fromJust rcv,
                            fromJust rdv, fromJust rev, fromJust rfv,
                            fromJust rgv, fromJust rhv, fromJust riv,
                            fromJust rjv, fromJust rkv) in
               if legal value
               then (Just value,(render errhexps,errstate))
               else (Nothing,   (renderError render errmsg errhexps,errstate)))


--- Widget for list structures where the list elements are vertically
--- aligned in a table.
wList :: WuiSpec a -> WuiSpec [a]
wList (WuiSpec rendera showa reada) =
  WuiSpec (renderList,"Illegal list:",const True)
          (\wparams vas ->
              listWidget (renderOf wparams) (unzip (map (showa rendera) vas)))
          (\ (render,errmsg,legal) env s ->
            let rvs = map (reada rendera env) (state2states s)
             in if Nothing `elem` (map fst rvs)
                then (Nothing, listWidget render (unzip (map snd rvs)))
                else let value = map (fromJust . fst) rvs in
                     if legal value
                     then (Just value, listWidget render (unzip (map snd rvs)))
                     else (Nothing, listWidget (renderError render errmsg)
                                               (unzip (map snd rvs))) )
 where
  listWidget render (hes,refs) = (render hes, states2state refs)

--- Add headings to a standard WUI for list structures:
wListWithHeadings :: [String] -> WuiSpec a -> WuiSpec [a]
wListWithHeadings headings wspec =
  wList wspec `withRendering` renderHeadings
 where
  renderHeadings hs = addHeadings (renderList hs) (map (\s->[htxt s]) headings)

--- Widget for list structures where the list elements are horizontally
--- aligned in a table.
wHList :: WuiSpec a -> WuiSpec [a]
wHList wspec = wList wspec `withRendering` renderTuple


--- WUI for matrices, i.e., list of list of elements
--- visualized as a matrix.
wMatrix :: WuiSpec a -> WuiSpec [[a]]
wMatrix wspec = wList (wHList wspec)


--- WUI for union types.
--- Here we provide only the implementation for Either types
--- since other types with more alternatives can be easily reduced to this case.
wEither :: WuiSpec a -> WuiSpec b -> WuiSpec (Either a b)
wEither (WuiSpec rendera showa reada) (WuiSpec renderb showb readb) =
 WuiSpec (head, "?", const True) showEither readEither
 where
  showEither wparams (Left va) =
    let (hea,rta) = showa rendera va
     in ((renderOf wparams) [hea], altstate2state (1,rta))
  showEither wparams (Right vb) =
    let (heb,rtb) = showb renderb vb
     in ((renderOf wparams) [heb], altstate2state (2,rtb))

  readEither (render,errmsg,legal) env s =
    let (altindex,rab) = state2altstate s
     in case altindex of
         1 -> let (rv,(he,rst)) = reada rendera env rab
               in checkValue (rv==Nothing) (Left (fromJust rv))
                             he (altstate2state(1,rst))
         2 -> let (rv,(he,rst)) = readb renderb env rab
               in checkValue (rv==Nothing) (Right (fromJust rv))
                             he (altstate2state(2,rst))
   where
     checkValue isnothing value hexp altstate =
        if isnothing
        then (Nothing, (render [hexp], altstate))
        else if legal value
             then (Just value, (render [hexp], altstate))
             else (Nothing,    (renderError render errmsg [hexp], altstate))


-------------------------------------------------------------------------------
-- Definition of standard rendering functions

--- standard rendering of tuples as a table with a single row:
renderTuple :: Rendering
renderTuple hexps = table [map (\h->[h]) hexps]

-- Standard error message for tuples:
tupleError = "Illegal combination:"

-- standard rendering of lists as a table with a row for each item:
renderList :: Rendering
renderList hexps = mergeTableOfTable (table (map (\h->[[h]]) hexps))
                                                `addAttr` ("border","1")

-- Combine a rendering with an error message.
-- The error message is put as the first row of a table with background color
-- yellow.
renderError :: Rendering -> String -> Rendering
renderError render errmsg hexps =
  table [[[boldRedTxt errmsg]], [[render hexps]]] 
                  `addAttr` ("bgcolor","#ffff00") -- background color: yellow

boldRedTxt s = HtmlStruct "font" [("color","#ff0000")] [bold [htxt s]]


mergeTableOfTable :: HtmlExp -> HtmlExp
mergeTableOfTable (HtmlStruct "table" attrs rows) =
  HtmlStruct "table" attrs
             (if all isRowWithSingleTableData rows
              then map mergeRowWithSingleTableData rows
              else rows )

isRowWithSingleTableData row = case row of
   (HtmlStruct "tr" []
        [HtmlStruct "td" []
            [HtmlStruct "table" _ [HtmlStruct "tr" _ _]]]) -> True
   _ -> False

mergeRowWithSingleTableData 
  (HtmlStruct "tr" [] [HtmlStruct "td" [] [HtmlStruct "table" _ [row]]]) = row


-------------------------------------------------------------------------------
-- Main operations to generate HTML structures and handlers from
-- WUI specifications:

--- Generates an HTML form from a WUI data specification,
--- an initial value and an update form.
mainWUI :: WuiSpec a -> a -> (a -> IO HtmlForm) -> HtmlForm
mainWUI wuispec val store =
  let (hexp,handler) = wui2html wuispec val store
   in form "WUI" [hexp, breakline, button "Submit" handler]

--- Generates HTML editors and a handler from a WUI data specification,
--- an initial value and an update form.
wui2html :: WuiSpec a -> a -> (a -> IO HtmlForm) -> (HtmlExp,HtmlHandler)
wui2html wspec val store = wuiWithErrorForm wspec val store standardErrorForm

--- A standard error form for WUIs.
standardErrorForm :: HtmlExp -> HtmlHandler -> HtmlForm
standardErrorForm hexp handler =
  standardForm "Input error" [hexp, button "Submit" handler]


--- Puts a WUI into a HTML form containing "holes" for the WUI and the
--- handler.
wuiInForm :: WuiSpec a -> a -> (a -> IO HtmlForm)
             -> (HtmlExp -> HtmlHandler -> HtmlForm) -> IO HtmlForm
wuiInForm wspec val store userform =
  let (hexp,handler) = wuiWithErrorForm wspec val store userform
   in return (userform hexp handler)


--- Generates HTML editors and a handler from a WUI data specification,
--- an initial value and an update form. In addition to wui2html,
--- we can provide a skeleton form used to show illegal inputs.
wuiWithErrorForm :: WuiSpec a -> a -> (a -> IO HtmlForm)
                    -> (HtmlExp -> HtmlHandler -> HtmlForm)
                    -> (HtmlExp,HtmlHandler)
wuiWithErrorForm wspec val store errorform =
        showAndReadWUI wspec store errorform (generateWUI wspec val)

generateWUI :: WuiSpec a -> a
                -> (HtmlExp, CgiEnv -> (Maybe a,(HtmlExp,WuiState)))
generateWUI (WuiSpec wparams showhtml readval) val =
  let (htmledits,wstate) = showhtml wparams val
   in (htmledits, \env -> readval wparams env wstate)

showAndReadWUI :: WuiSpec a -> (a -> IO HtmlForm)
                            -> (HtmlExp -> HtmlHandler -> HtmlForm)
                            -> (HtmlExp,CgiEnv -> (Maybe a,(HtmlExp,WuiState)))
                            -> (HtmlExp,HtmlHandler)
showAndReadWUI wspec@(WuiSpec wparams _ readval) store errorform
               (htmledits,readenv) =
  (htmledits,
   \env ->
     let (mbnewval, (htmlerrform,errwstate)) = readenv env in
     maybe (let (errhexp,errhdl) =
                   showAndReadWUI wspec
                                  store
                                  errorform
                                  (htmlerrform,
                                   \errenv -> readval wparams errenv errwstate)
            in return (errorform errhexp errhdl))
           (\newval -> seq (normalForm newval) -- to strip off unused lvars
                           (store newval))
           mbnewval)


--------------------------------------------------------------------------
